# Learning Across Decentralized Multi-Modal Remote Sensing Archives with Federated Learning
This repository contains the code of the paper [Learning Across Decentralized Multi-Modal Remote Sensing Archives with Federated Learning](https://arxiv.org/abs/2306.00792) accepted at IEEE International Geoscience and Remote Sensing Symposium (IGARSS) 2023. This work has been done at the [Remote Sensing Image Analysis group](https://www.rsim.tu-berlin.de/menue/remote_sensing_image_analysis_group/) by [Baris Buyuktas](https://rsim.berlin/team/members/baris-buyuktas), [Gencer Sumbul](http://www.user.tu-berlin.de/gencersumbul/), and [Begüm Demir](https://begumdemir.com/).

If you use this code, please cite our paper given below:

> B. Buyuktas, G. Sumbul and B. Demіr, "Learning Across Decentralized Multi-Modal Remote Sensing Archives with Federated Learning", IEEE International Geoscience and Remote Sensing Symposium (IGARSS), 2023.

```
@inproceedings{MM-FL,
  author={B. {Buyuktas} and G. {Sumbul} and B. {Demіr}},
  booktitle={IEEE International Geoscience and Remote Sensing Symposium (IGARSS)}, 
  title={Learning Across Decentralized Multi-Modal Remote Sensing Archives with Federated Learning}, 
  year={2023}
}

```

This repository contains code (in parts) that has been adapted from:
* https://github.com/ki-ljl/FedProx-PyTorch
* https://github.com/roysubhankar/dwt-domain-adaptation


## Introduction
The development of federated learning (FL) methods, which aim to learn from distributed databases (i.e., clients) without accessing data on clients, has recently attracted great attention. Most of these methods assume that the clients are associated with the same data modality. However, remote sensing (RS) images in different clients can be associated with different data modalities that can improve the classification performance when jointly used. To address this problem, in this paper we introduce a novel multi-modal FL framework that aims to learn from decentralized multi-modal RS image archives for RS image classification problems. The proposed framework is made up of three modules: 1) multi-modal fusion (MF); 2) feature whitening (FW); and 3) mutual information maximization (MIM). The MF module performs iterative model averaging to learn without accessing data on clients in the case that clients are associated with different data modalities. The FW module aligns the representations learned among the different clients. The MIM module maximizes the similarity of images from different modalities. Experimental results show the effectiveness of the proposed framework compared to iterative model averaging, which is a widely used algorithm in FL.

## Prerequisites
The code in this repository requires Python 3.10.4, pytorch 1.12.1. 

The code is tested in Ubuntu 18.04. 

An exemplary setup which contains everything needed:
```
(1) wget  https://repo.continuum.io/miniconda/Miniconda3-latest-Linux-x86_64.sh
(2) bash Miniconda3-latest-Linux-x86_64.sh (say yes to append path to bashrc)
(3) source .bashrc
(4) sudo add-apt-repository ppa:ubuntugis/ppa && sudo apt-get update && sudo apt-get install g++
(5) sudo apt-get install libgdal-dev gdal-bin
(6) ogrinfo --version
(7) conda activate base
(8) conda install matplotlib scipy scikit-learn scikit-image tqdm pillow pytorch  

```


## Training Arguments

* `epoch`: number of local training
* `round`: number of communication rounds
* `clients`: number of total clients
* `lr`: learning rate
* `batch`: local batch size
* `decay`: weight decay
* `optimizer`: type of optimizer
* `sampling`: sampling rate
* `step_size`: step size
* `gamma`: learning rate decay per global round

## How to run

`python main.py`


## Authors
**Baris Buyuktas**
https://rsim.berlin/team/members/baris-buyuktas

**Gencer Sümbül**
http://www.user.tu-berlin.de/gencersumbul



## License
The code in this repository to facilitate the use of the `Learning Across Decentralized Multi-Modal Remote Sensing Archives with Federated Learning` is licensed under the **MIT License**:

```
MIT License

Copyright (c) 2023 The Authors of The Paper, "Learning Across Decentralized Multi-Modal Remote Sensing Archives with Federated Learning"

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

