
from torch import nn
import torch.nn.init as init
from torchvision import models
import torch
torch.cuda.empty_cache()

import whitening


class whitening_scale_shift(nn.Module):
    # copied from https://github.com/roysubhankar/dwt-domain-adaptation/blob/9faac19b13ec89f4d4446d816e3e132df4eb012a/resnet50_dwt_mec_officehome.py#L40 
	def __init__(self, planes, group_size, running_mean, running_variance, track_running_stats=True, affine=True):
		super(whitening_scale_shift, self).__init__()
		self.planes = planes
		self.group_size = group_size
		self.track_running_stats = track_running_stats
		self.affine = affine
		self.running_mean = running_mean
		self.running_variance = running_variance
        
		self.wh = whitening.WTransform2d(self.planes, 
										 self.group_size, 
										 running_m=self.running_mean, 
										 running_var=self.running_variance, 
										 track_running_stats=self.track_running_stats)
		if self.affine:
			self.gamma = nn.Parameter(torch.ones(self.planes, 1, 1))
			self.beta = nn.Parameter(torch.zeros(self.planes, 1, 1))

	def forward(self, x):
		out = self.wh(x)
		if self.affine:
			out = out * self.gamma + self.beta
		return out

        
class ANN(nn.Module):
    def __init__(self,args, name, num_cls=19, channels=2, FC_dim=2048, pretrained=True):
        super(ANN, self).__init__()
        self.name = name
        self.len = 0
        self.loss = 0

        resnet = models.resnet50(pretrained=pretrained)
        resnet.layer1[0].bn1 = whitening_scale_shift(planes=64, group_size=4,running_mean=None, running_variance=None)
        resnet.layer1[0].bn2 = whitening_scale_shift(planes=64, group_size=4,running_mean=None, running_variance=None)
        resnet.layer1[0].bn3 = whitening_scale_shift(planes=256, group_size=4,running_mean=None, running_variance=None)
        resnet.layer1[0].downsample[1] = whitening_scale_shift(planes=256, group_size=4,running_mean=None, running_variance=None)
        resnet.layer1[1].bn1 = whitening_scale_shift(planes=64, group_size=4,running_mean=None, running_variance=None)
        resnet.layer1[1].bn2 = whitening_scale_shift(planes=64, group_size=4,running_mean=None, running_variance=None)
        resnet.layer1[1].bn3 = whitening_scale_shift(planes=256, group_size=4,running_mean=None, running_variance=None)
        resnet.layer1[2].bn1 = whitening_scale_shift(planes=64, group_size=4,running_mean=None, running_variance=None)
        resnet.layer1[2].bn2 = whitening_scale_shift(planes=64, group_size=4,running_mean=None, running_variance=None)
        resnet.layer1[2].bn3 = whitening_scale_shift(planes=256, group_size=4,running_mean=None, running_variance=None)

            # layer 2
        resnet.layer2[0].bn1 = whitening_scale_shift(planes=128, group_size=4,running_mean=None, running_variance=None)
        resnet.layer2[0].bn2 = whitening_scale_shift(planes=128, group_size=4,running_mean=None, running_variance=None)
        resnet.layer2[0].bn3 = whitening_scale_shift(planes=512, group_size=4,running_mean=None, running_variance=None)
        resnet.layer2[0].downsample[1] = whitening_scale_shift(planes=512, group_size=4,running_mean=None, running_variance=None)
        resnet.layer2[1].bn1 = whitening_scale_shift(planes=128, group_size=4,running_mean=None, running_variance=None)
        resnet.layer2[1].bn2 = whitening_scale_shift(planes=128, group_size=4,running_mean=None, running_variance=None)      
        resnet.layer2[1].bn3 = whitening_scale_shift(planes=512, group_size=4,running_mean=None, running_variance=None)
        resnet.layer2[2].bn1 = whitening_scale_shift(planes=128, group_size=4,running_mean=None, running_variance=None)
        resnet.layer2[2].bn2 = whitening_scale_shift(planes=128, group_size=4,running_mean=None, running_variance=None)      
        resnet.layer2[2].bn3 = whitening_scale_shift(planes=512, group_size=4,running_mean=None, running_variance=None)
        resnet.layer2[3].bn1 = whitening_scale_shift(planes=128, group_size=4,running_mean=None, running_variance=None)
        resnet.layer2[3].bn2 = whitening_scale_shift(planes=128, group_size=4,running_mean=None, running_variance=None)      
        resnet.layer2[3].bn3 = whitening_scale_shift(planes=512, group_size=4,running_mean=None, running_variance=None)
  
            # layer 3
        resnet.layer3[0].bn1 = whitening_scale_shift(planes=256, group_size=4,running_mean=None, running_variance=None)
        resnet.layer3[0].bn2 = whitening_scale_shift(planes=256, group_size=4,running_mean=None, running_variance=None)
        resnet.layer3[0].bn3 = whitening_scale_shift(planes=1024, group_size=4,running_mean=None, running_variance=None)
        resnet.layer3[0].downsample[1] = whitening_scale_shift(planes=1024, group_size=4,running_mean=None, running_variance=None)
        resnet.layer3[1].bn1 = whitening_scale_shift(planes=256, group_size=4,running_mean=None, running_variance=None)
        resnet.layer3[1].bn2 = whitening_scale_shift(planes=256, group_size=4,running_mean=None, running_variance=None)
        resnet.layer3[1].bn3 = whitening_scale_shift(planes=1024, group_size=4,running_mean=None, running_variance=None)
        resnet.layer3[2].bn1 = whitening_scale_shift(planes=256, group_size=4,running_mean=None, running_variance=None)
        resnet.layer3[2].bn2 = whitening_scale_shift(planes=256, group_size=4,running_mean=None, running_variance=None)
        resnet.layer3[2].bn3 = whitening_scale_shift(planes=1024, group_size=4,running_mean=None, running_variance=None)
        resnet.layer3[3].bn1 = whitening_scale_shift(planes=256, group_size=4,running_mean=None, running_variance=None)
        resnet.layer3[3].bn2 = whitening_scale_shift(planes=256, group_size=4,running_mean=None, running_variance=None)
        resnet.layer3[3].bn3 = whitening_scale_shift(planes=1024, group_size=4,running_mean=None, running_variance=None)
        resnet.layer3[4].bn1 = whitening_scale_shift(planes=256, group_size=4,running_mean=None, running_variance=None)
        resnet.layer3[4].bn2 = whitening_scale_shift(planes=256, group_size=4,running_mean=None, running_variance=None)
        resnet.layer3[4].bn3 = whitening_scale_shift(planes=1024, group_size=4,running_mean=None, running_variance=None)
        resnet.layer3[5].bn1 = whitening_scale_shift(planes=256, group_size=4,running_mean=None, running_variance=None)
        resnet.layer3[5].bn2 = whitening_scale_shift(planes=256, group_size=4,running_mean=None, running_variance=None)
        resnet.layer3[5].bn3 = whitening_scale_shift(planes=1024, group_size=4,running_mean=None, running_variance=None)
            # layer 4
        resnet.layer4[0].bn1 = whitening_scale_shift(planes=512, group_size=4,running_mean=None, running_variance=None)
        resnet.layer4[0].bn2 = whitening_scale_shift(planes=512, group_size=4,running_mean=None, running_variance=None)
        resnet.layer4[0].bn3 = whitening_scale_shift(planes=2048, group_size=4,running_mean=None, running_variance=None)
        resnet.layer4[0].downsample[1] = whitening_scale_shift(planes=2048, group_size=4,running_mean=None, running_variance=None)
        resnet.layer4[1].bn1 = whitening_scale_shift(planes=512, group_size=4,running_mean=None, running_variance=None)
        resnet.layer4[1].bn2 = whitening_scale_shift(planes=512, group_size=4,running_mean=None, running_variance=None)
        resnet.layer4[1].bn3 = whitening_scale_shift(planes=2048, group_size=4,running_mean=None, running_variance=None)
        resnet.layer4[2].bn1 = whitening_scale_shift(planes=512, group_size=4,running_mean=None, running_variance=None)
        resnet.layer4[2].bn2 = whitening_scale_shift(planes=512, group_size=4,running_mean=None, running_variance=None)
        resnet.layer4[2].bn3 = whitening_scale_shift(planes=2048, group_size=4,running_mean=None, running_variance=None)
        self.conv1 = nn.Conv2d(channels, 64, kernel_size=(7, 7), stride=(2, 2), padding=(3, 3), bias=False)
        self.encoder = nn.Sequential(
            self.conv1,
            resnet.bn1,
            resnet.relu,
            resnet.maxpool,
            resnet.layer1,
            resnet.layer2,
            resnet.layer3,
            resnet.layer4,
            resnet.avgpool
        )
        
        self.FC = nn.Linear(FC_dim*2, num_cls)
       
        
        self.apply(weights_init_kaiming)
        self.apply(fc_init_weights)

    def forward(self, x):
        x = self.encoder(x)
      #  x=self.bn(x)
        x = x.view(x.size(0), -1)
        conc = torch.zeros(x.shape[0],2048).cuda()
        x = torch.cat((x,conc),1)
        logits = self.FC(x)

        return logits

class ANN_S2(nn.Module):
    def __init__(self,args, name, num_cls=19, channels=10, FC_dim=2048, pretrained=True):
        super(ANN_S2, self).__init__()
        self.name = name
        self.len = 0
        self.loss = 0

        
        resnet = models.resnet50(pretrained=pretrained)
        resnet.layer1[0].bn1 = whitening_scale_shift(planes=64, group_size=4,running_mean=None, running_variance=None)
        resnet.layer1[0].bn2 = whitening_scale_shift(planes=64, group_size=4,running_mean=None, running_variance=None)
        resnet.layer1[0].bn3 = whitening_scale_shift(planes=256, group_size=4,running_mean=None, running_variance=None)
        resnet.layer1[0].downsample[1] = whitening_scale_shift(planes=256, group_size=4,running_mean=None, running_variance=None)
        resnet.layer1[1].bn1 = whitening_scale_shift(planes=64, group_size=4,running_mean=None, running_variance=None)
        resnet.layer1[1].bn2 = whitening_scale_shift(planes=64, group_size=4,running_mean=None, running_variance=None)
        resnet.layer1[1].bn3 = whitening_scale_shift(planes=256, group_size=4,running_mean=None, running_variance=None)
        resnet.layer1[2].bn1 = whitening_scale_shift(planes=64, group_size=4,running_mean=None, running_variance=None)
        resnet.layer1[2].bn2 = whitening_scale_shift(planes=64, group_size=4,running_mean=None, running_variance=None)
        resnet.layer1[2].bn3 = whitening_scale_shift(planes=256, group_size=4,running_mean=None, running_variance=None)

            # layer 2
        resnet.layer2[0].bn1 = whitening_scale_shift(planes=128, group_size=4,running_mean=None, running_variance=None)
        resnet.layer2[0].bn2 = whitening_scale_shift(planes=128, group_size=4,running_mean=None, running_variance=None)
        resnet.layer2[0].bn3 = whitening_scale_shift(planes=512, group_size=4,running_mean=None, running_variance=None)
        resnet.layer2[0].downsample[1] = whitening_scale_shift(planes=512, group_size=4,running_mean=None, running_variance=None)
        resnet.layer2[1].bn1 = whitening_scale_shift(planes=128, group_size=4,running_mean=None, running_variance=None)
        resnet.layer2[1].bn2 = whitening_scale_shift(planes=128, group_size=4,running_mean=None, running_variance=None)      
        resnet.layer2[1].bn3 = whitening_scale_shift(planes=512, group_size=4,running_mean=None, running_variance=None)
        resnet.layer2[2].bn1 = whitening_scale_shift(planes=128, group_size=4,running_mean=None, running_variance=None)
        resnet.layer2[2].bn2 = whitening_scale_shift(planes=128, group_size=4,running_mean=None, running_variance=None)      
        resnet.layer2[2].bn3 = whitening_scale_shift(planes=512, group_size=4,running_mean=None, running_variance=None)
        resnet.layer2[3].bn1 = whitening_scale_shift(planes=128, group_size=4,running_mean=None, running_variance=None)
        resnet.layer2[3].bn2 = whitening_scale_shift(planes=128, group_size=4,running_mean=None, running_variance=None)      
        resnet.layer2[3].bn3 = whitening_scale_shift(planes=512, group_size=4,running_mean=None, running_variance=None)
  
            # layer 3
        resnet.layer3[0].bn1 = whitening_scale_shift(planes=256, group_size=4,running_mean=None, running_variance=None)
        resnet.layer3[0].bn2 = whitening_scale_shift(planes=256, group_size=4,running_mean=None, running_variance=None)
        resnet.layer3[0].bn3 = whitening_scale_shift(planes=1024, group_size=4,running_mean=None, running_variance=None)
        resnet.layer3[0].downsample[1] = whitening_scale_shift(planes=1024, group_size=4,running_mean=None, running_variance=None)
        resnet.layer3[1].bn1 = whitening_scale_shift(planes=256, group_size=4,running_mean=None, running_variance=None)
        resnet.layer3[1].bn2 = whitening_scale_shift(planes=256, group_size=4,running_mean=None, running_variance=None)
        resnet.layer3[1].bn3 = whitening_scale_shift(planes=1024, group_size=4,running_mean=None, running_variance=None)
        resnet.layer3[2].bn1 = whitening_scale_shift(planes=256, group_size=4,running_mean=None, running_variance=None)
        resnet.layer3[2].bn2 = whitening_scale_shift(planes=256, group_size=4,running_mean=None, running_variance=None)
        resnet.layer3[2].bn3 = whitening_scale_shift(planes=1024, group_size=4,running_mean=None, running_variance=None)
        resnet.layer3[3].bn1 = whitening_scale_shift(planes=256, group_size=4,running_mean=None, running_variance=None)
        resnet.layer3[3].bn2 = whitening_scale_shift(planes=256, group_size=4,running_mean=None, running_variance=None)
        resnet.layer3[3].bn3 = whitening_scale_shift(planes=1024, group_size=4,running_mean=None, running_variance=None)
        resnet.layer3[4].bn1 = whitening_scale_shift(planes=256, group_size=4,running_mean=None, running_variance=None)
        resnet.layer3[4].bn2 = whitening_scale_shift(planes=256, group_size=4,running_mean=None, running_variance=None)
        resnet.layer3[4].bn3 = whitening_scale_shift(planes=1024, group_size=4,running_mean=None, running_variance=None)
        resnet.layer3[5].bn1 = whitening_scale_shift(planes=256, group_size=4,running_mean=None, running_variance=None)
        resnet.layer3[5].bn2 = whitening_scale_shift(planes=256, group_size=4,running_mean=None, running_variance=None)
        resnet.layer3[5].bn3 = whitening_scale_shift(planes=1024, group_size=4,running_mean=None, running_variance=None)
            # layer 4
        resnet.layer4[0].bn1 = whitening_scale_shift(planes=512, group_size=4,running_mean=None, running_variance=None)
        resnet.layer4[0].bn2 = whitening_scale_shift(planes=512, group_size=4,running_mean=None, running_variance=None)
        resnet.layer4[0].bn3 = whitening_scale_shift(planes=2048, group_size=4,running_mean=None, running_variance=None)
        resnet.layer4[0].downsample[1] = whitening_scale_shift(planes=2048, group_size=4,running_mean=None, running_variance=None)
        resnet.layer4[1].bn1 = whitening_scale_shift(planes=512, group_size=4,running_mean=None, running_variance=None)
        resnet.layer4[1].bn2 = whitening_scale_shift(planes=512, group_size=4,running_mean=None, running_variance=None)
        resnet.layer4[1].bn3 = whitening_scale_shift(planes=2048, group_size=4,running_mean=None, running_variance=None)
        resnet.layer4[2].bn1 = whitening_scale_shift(planes=512, group_size=4,running_mean=None, running_variance=None)
        resnet.layer4[2].bn2 = whitening_scale_shift(planes=512, group_size=4,running_mean=None, running_variance=None)
        resnet.layer4[2].bn3 = whitening_scale_shift(planes=2048, group_size=4,running_mean=None, running_variance=None)
        self.conv1 = nn.Conv2d(channels, 64, kernel_size=(7, 7), stride=(2, 2), padding=(3, 3), bias=False)
        self.encoder = nn.Sequential(
            
            self.conv1,
            resnet.bn1,
            resnet.relu,
            resnet.maxpool,
            resnet.layer1,
            resnet.layer2,
            resnet.layer3,
            resnet.layer4,
            resnet.avgpool
        )
   
        self.FC = nn.Linear(FC_dim*2, num_cls)
        self.apply(weights_init_kaiming)
        self.apply(fc_init_weights)
    
    def forward(self, x):
        x = self.encoder(x)
      
        x = x.view(x.size(0), -1)
       
        conc = torch.zeros(x.shape[0],2048).cuda()
        x = torch.cat((conc,x),1)
        logits = self.FC(x)

        return logits


def weights_init_kaiming(m):
        classname = m.__class__.__name__
        if classname.find('Conv2d') != -1:
            init.kaiming_normal_(m.weight.data)


def fc_init_weights(m):
            if type(m) == nn.Linear:
                    init.kaiming_normal_(m.weight.data)


