
import copy
import random

import numpy as np
import torch
from tqdm import tqdm

from model import ANN
from model import ANN_S2
from client import train, test


class Fed:
    def __init__(self, args):
        self.args = args
        self.nn = ANN(args=self.args, name='server').to(args.device)
        self.nn_s2 = ANN_S2(args=self.args, name='server1').to(args.device)
        self.nn_ultimate=0
        self.nns = []
        self.s1=ANN(args=self.args, name='server2').to(args.device)
        self.s2=ANN_S2(args=self.args, name='server3').to(args.device)
        for i in range(3):
            temp = copy.deepcopy(self.nn_s2)
            temp.name = self.args.clients[i]
            self.nns.append(temp)
        for i in range(3,6):
            temp = copy.deepcopy(self.nn)
            temp.name = self.args.clients[i]
            self.nns.append(temp)

    def server(self):
        for t in tqdm(range(self.args.r)):
            print('round', t + 1, ':')
            index=[0,1,2,3,4,5]
            self.client_update(index)
            self.aggregation(index)
            self.aggregation_input_s1(index)
            self.aggregation_input_s2(index)
            
            torch.save(self.nn.state_dict(), '/data/scenario2-1.pth')
            torch.save(self.s1.state_dict(), '/data/scenario2-2.pth')
            torch.save(self.s2.state_dict(), '/data/scenario2-3.pth')
          
            print(test(self.args, self.nn))
            
        return self.nn

    def manipulation(model,layer,parameters):
        for name, param in model.named_parameters():
          if param.requires_grad:
            if name == layer:
               param.data = parameters
  
        
    def aggregation(self, index):
        
        ult_models=[]
       
        for i in index:
              ult_models.append(self.nns[i].state_dict())
        
        nums=[]
        for j in index:
            
            nums.append( self.nns[j].len )

        
        nums = [x / 10 for x in nums]
    
        sdB=copy.deepcopy(ult_models[0])
        sdA=copy.deepcopy(ult_models[0])
        
        first=nums[0]  
        for i in range(len(ult_models)-1):
           tmp=copy.deepcopy(ult_models[i+1])
           for enu,key in enumerate(sdA):
             if(enu>1):
                sdB[key] = (tmp[key]*nums[i+1] + sdA[key]*first)     
           sdA=copy.deepcopy(sdB)
           first=1
        

        for key in sdB:
            sdB[key] = (sdB[key]) / sum(nums)
            
        sdB['conv1.weight']=(ult_models[3]['conv1.weight']*nums[3]+ult_models[4]['conv1.weight']*nums[4]+ult_models[5]['conv1.weight']*nums[5])/(nums[3]+nums[4]+nums[5])
        sdB['encoder.0.weight']=(ult_models[3]['encoder.0.weight']*nums[3]+ult_models[4]['encoder.0.weight']*nums[4]+ult_models[5]['encoder.0.weight']*nums[5])/(nums[3]+nums[4]+nums[5])

                    
     
        self.nn.load_state_dict(sdB)
       
      
    
    def aggregation_input_s2(self, index):
        #
        ult_models=[]
        index=np.arange(0,3)
        for i in index:
              ult_models.append(self.nns[i].state_dict())
        
        nums=[]
        for j in index:
            # normal
            nums.append( self.nns[j].len )
        s = 0
        for j in index:
            
            s += self.nns[j].len

        nums = [x / 10 for x in nums]
    
        sdB=copy.deepcopy(ult_models[0])
        sdA=copy.deepcopy(ult_models[0])
        
        first=nums[0]         
        
        for i in range(len(ult_models)-1):
           tmp=copy.deepcopy(ult_models[i+1])
           for key in sdA:
             sdB[key] = (tmp[key]*nums[i+1] + sdA[key]*first)
           sdA=copy.deepcopy(sdB)
           first=1
        

        for key in sdB:
            sdB[key] = (sdB[key]) / sum(nums)  
     
        self.s2.load_state_dict(sdB)

    def aggregation_input_s1(self, index):
       
        ult_models=[]
        index=np.arange(3,6)
        for i in index:
              ult_models.append(self.nns[i].state_dict())
        
        nums=[]
        for j in index:
            
            nums.append( self.nns[j].len )
        s = 0
        for j in index:
            
            s += self.nns[j].len

        nums = [x / 10 for x in nums]
        
        sdB=copy.deepcopy(ult_models[0])
        sdA=copy.deepcopy(ult_models[0])
        
        first=nums[0]         
        
        for i in range(len(ult_models)-1):
           tmp=copy.deepcopy(ult_models[i+1])
           for key in sdA:
             sdB[key] = (tmp[key]*nums[i+1] + sdA[key]*first)
           sdA=copy.deepcopy(sdB)
           first=1
        

        for key in sdB:
            sdB[key] = (sdB[key]) / sum(nums)  
     
        self.s1.load_state_dict(sdB)
    
  
    def dispatch(self, index):
        for j in index:
            for old_params, new_params in zip(self.nns[j].parameters(), self.nn.parameters()):
                old_params.data = new_params.data.clone()

    


     
    def client_update(self, index):  # update nn
        for k in index:
        
            self.nns[k] = train(self.args, self.nns[k], self.nn,k,self.s1,self.s2)
    def global_test(self):
        model = self.nn_s2
        model.eval()
 
        test(self.args, model)
        
       